package dev.fyloz.colorrecipesexplorer.model

import com.fasterxml.jackson.annotation.JsonIgnore
import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.exception.CannotDeleteException
import dev.fyloz.colorrecipesexplorer.exception.NotFoundException
import dev.fyloz.colorrecipesexplorer.model.validation.NullOrNotBlank
import dev.fyloz.colorrecipesexplorer.model.validation.NullOrSize
import org.springframework.web.multipart.MultipartFile
import java.net.URI
import javax.persistence.*
import javax.validation.constraints.Min
import javax.validation.constraints.NotBlank
import javax.validation.constraints.NotNull

private const val MATERIAL_ID_NULL_MESSAGE = "Un identifiant est requis"
private const val MATERIAL_NAME_NULL_MESSAGE = "Un nom est requis"
private const val MATERIAL_INVENTORY_QUANTITY_NULL_MESSAGE = "Une quantité est requise"
private const val MATERIAL_INVENTORY_QUANTITY_NEGATIVE_MESSAGE = "La quantité doit être supérieure ou égale à 0"
private const val MATERIAL_TYPE_NULL_MESSAGE = "Un type de produit est requis"

private const val MATERIAL_QUANTITY_MATERIAL_NULL_MESSAGE = "Un produit est requis"
private const val MATERIAL_QUANTITY_QUANTITY_NULL_MESSAGE = "Une quantité est requises"
private const val MATERIAL_QUANTITY_QUANTITY_NEGATIVE_MESSAGE = "La quantité doit être supérieure ou égale à 0"

const val SIMDUT_FILES_PATH = "pdf/simdut"

@Entity
@Table(name = "material")
data class Material(
    @Id
    @GeneratedValue(strategy = GenerationType.IDENTITY)
    override val id: Long?,

    @Column(unique = true)
    override var name: String,

    @Column(name = "inventory_quantity")
    var inventoryQuantity: Float,

    @Column(name = "mix_type")
    val isMixType: Boolean,

    @ManyToOne
    @JoinColumn(name = "material_type_id")
    var materialType: MaterialType?
) : NamedModel {
    val simdutFilePath
        @JsonIgnore
        @Transient
        get() = "$SIMDUT_FILES_PATH/$name.pdf"
}

open class MaterialSaveDto(
    @field:NotBlank(message = MATERIAL_NAME_NULL_MESSAGE)
    val name: String,

    @field:NotNull(message = MATERIAL_INVENTORY_QUANTITY_NULL_MESSAGE)
    @field:Min(value = 0, message = MATERIAL_INVENTORY_QUANTITY_NEGATIVE_MESSAGE)
    val inventoryQuantity: Float,

    @field:NotNull(message = MATERIAL_TYPE_NULL_MESSAGE)
    val materialTypeId: Long,

    val simdutFile: MultipartFile? = null
) : EntityDto<Material>

open class MaterialUpdateDto(
    @field:NotNull(message = MATERIAL_ID_NULL_MESSAGE)
    val id: Long,

    @field:NullOrNotBlank(message = MATERIAL_NAME_NULL_MESSAGE)
    val name: String?,

    @field:NullOrSize(min = 0, message = MATERIAL_INVENTORY_QUANTITY_NEGATIVE_MESSAGE)
    val inventoryQuantity: Float?,

    val materialTypeId: Long?,

    val simdutFile: MultipartFile? = null
) : EntityDto<Material>

data class MaterialOutputDto(
    val id: Long,
    val name: String,
    val inventoryQuantity: Float,
    val isMixType: Boolean,
    val materialType: MaterialType,
    val simdutUrl: String?
)

data class MaterialQuantityDto(
    @field:NotNull(message = MATERIAL_QUANTITY_MATERIAL_NULL_MESSAGE)
    val material: Long,

    @field:NotNull(message = MATERIAL_QUANTITY_QUANTITY_NULL_MESSAGE)
    @field:Min(value = 0, message = MATERIAL_QUANTITY_QUANTITY_NEGATIVE_MESSAGE)
    val quantity: Float
)

// === DSL ===

fun material(
    id: Long? = null,
    name: String = "name",
    inventoryQuantity: Float = 0f,
    isMixType: Boolean = false,
    materialType: MaterialType? = materialType(),
    op: Material.() -> Unit = {}
) = Material(id, name, inventoryQuantity, isMixType, materialType).apply(op)

fun material(
    material: Material,
    id: Long? = null,
    name: String? = null,
) = Material(
    id ?: material.id, name
        ?: material.name, material.inventoryQuantity, material.isMixType, material.materialType
)

fun materialSaveDto(
    name: String = "name",
    inventoryQuantity: Float = 0f,
    materialTypeId: Long = 0L,
    simdutFile: MultipartFile? = null,
    op: MaterialSaveDto.() -> Unit = {}
) = MaterialSaveDto(name, inventoryQuantity, materialTypeId, simdutFile).apply(op)

fun materialUpdateDto(
    id: Long = 0L,
    name: String? = "name",
    inventoryQuantity: Float? = 0f,
    materialTypeId: Long? = 0L,
    simdutFile: MultipartFile? = null,
    op: MaterialUpdateDto.() -> Unit = {}
) = MaterialUpdateDto(id, name, inventoryQuantity, materialTypeId, simdutFile).apply(op)

fun materialOutputDto(
    material: Material,
    simdutUrl: String?,
    op: MaterialOutputDto.() -> Unit = {}
) = MaterialOutputDto(
    id = material.id!!,
    name = material.name,
    inventoryQuantity = material.inventoryQuantity,
    isMixType = material.isMixType,
    materialType = material.materialType!!,
    simdutUrl = simdutUrl
).apply(op)

fun materialQuantityDto(
    materialId: Long,
    quantity: Float,
    op: MaterialQuantityDto.() -> Unit = {}
) = MaterialQuantityDto(materialId, quantity).apply(op)

// ==== Exceptions ====
private const val MATERIAL_NOT_FOUND_EXCEPTION_TITLE = "Material not found"
private const val MATERIAL_ALREADY_EXISTS_EXCEPTION_TITLE = "Material already exists"
private const val MATERIAL_CANNOT_DELETE_EXCEPTION_TITLE = "Cannot delete material"
private const val MATERIAL_EXCEPTION_ERROR_CODE = "material"

fun materialIdNotFoundException(id: Long) =
    NotFoundException(
        MATERIAL_EXCEPTION_ERROR_CODE,
        MATERIAL_NOT_FOUND_EXCEPTION_TITLE,
        "A material with the id $id could not be found",
        id
    )

fun materialNameNotFoundException(name: String) =
    NotFoundException(
        MATERIAL_EXCEPTION_ERROR_CODE,
        MATERIAL_NOT_FOUND_EXCEPTION_TITLE,
        "A material with the name $name could not be found",
        name,
        "name"
    )

fun materialIdAlreadyExistsException(id: Long) =
    AlreadyExistsException(
        MATERIAL_EXCEPTION_ERROR_CODE,
        MATERIAL_ALREADY_EXISTS_EXCEPTION_TITLE,
        "A material with the id $id already exists",
        id
    )

fun materialNameAlreadyExistsException(name: String) =
    AlreadyExistsException(
        MATERIAL_EXCEPTION_ERROR_CODE,
        MATERIAL_ALREADY_EXISTS_EXCEPTION_TITLE,
        "A material with the name $name already exists",
        name,
        "name"
    )

fun cannotDeleteMaterialException(material: Material) =
    CannotDeleteException(
        MATERIAL_EXCEPTION_ERROR_CODE,
        MATERIAL_CANNOT_DELETE_EXCEPTION_TITLE,
        "Cannot delete the material ${material.name} because one or more recipes depends on it"
    )
