package dev.fyloz.colorrecipesexplorer.rest

import dev.fyloz.colorrecipesexplorer.config.annotations.PreAuthorizeEditRecipes
import dev.fyloz.colorrecipesexplorer.config.annotations.PreAuthorizeRemoveRecipes
import dev.fyloz.colorrecipesexplorer.config.annotations.PreAuthorizeViewRecipes
import dev.fyloz.colorrecipesexplorer.config.properties.CreProperties
import dev.fyloz.colorrecipesexplorer.model.*
import dev.fyloz.colorrecipesexplorer.rest.files.FILE_CONTROLLER_PATH
import dev.fyloz.colorrecipesexplorer.service.MixService
import dev.fyloz.colorrecipesexplorer.service.RecipeImageService
import dev.fyloz.colorrecipesexplorer.service.RecipeService
import org.springframework.http.MediaType
import org.springframework.http.ResponseEntity
import org.springframework.security.access.prepost.PreAuthorize
import org.springframework.web.bind.annotation.*
import org.springframework.web.multipart.MultipartFile
import java.net.URLEncoder
import java.nio.charset.StandardCharsets
import javax.validation.Valid


private const val RECIPE_CONTROLLER_PATH = "api/recipe"
private const val MIX_CONTROLLER_PATH = "api/recipe/mix"

@RestController
@RequestMapping(RECIPE_CONTROLLER_PATH)
@PreAuthorizeViewRecipes
class RecipeController(
    private val recipeService: RecipeService,
    private val recipeImageService: RecipeImageService,
    private val creProperties: CreProperties
) {
    @GetMapping
    fun getAll() =
        ok(recipeService.getAll())

    @GetMapping("{id}")
    fun getById(@PathVariable id: Long) =
        ok(recipeService.getById(id))

    @PostMapping
    @PreAuthorizeEditRecipes
    fun save(@Valid @RequestBody recipe: RecipeSaveDto) =
        created<Recipe>(RECIPE_CONTROLLER_PATH) {
            recipeService.save(recipe)
        }

    @PutMapping
    @PreAuthorizeEditRecipes
    fun update(@Valid @RequestBody recipe: RecipeUpdateDto) =
        noContent {
            recipeService.update(recipe)
        }

    @PutMapping("public")
    @PreAuthorize("hasAuthority('EDIT_RECIPES_PUBLIC_DATA')")
    fun updatePublicData(@Valid @RequestBody publicDataDto: RecipePublicDataDto) =
        noContent {
            recipeService.updatePublicData(publicDataDto)
        }

    @DeleteMapping("{id}")
    @PreAuthorizeRemoveRecipes
    fun deleteById(@PathVariable id: Long) =
        noContent {
            recipeService.deleteById(id)
        }

    @PutMapping("{recipeId}/image", consumes = [MediaType.APPLICATION_OCTET_STREAM_VALUE])
    @PreAuthorizeEditRecipes
    fun downloadImage(@PathVariable recipeId: Long, image: MultipartFile): ResponseEntity<RecipeOutputDto> {
        recipeImageService.download(image, recipeService.getById(recipeId))
        return getById(recipeId)
    }

    @DeleteMapping("{recipeId}/image/{name}")
    @PreAuthorizeEditRecipes
    fun deleteImage(@PathVariable recipeId: Long, @PathVariable name: String) =
        noContent {
            recipeImageService.delete(recipeService.getById(recipeId), name)
        }

    private fun ok(recipe: Recipe) =
        ok(recipe.toOutput())

    private fun ok(recipes: Collection<Recipe>) =
        ok(recipes.map { it.toOutput() })

    private fun Recipe.toOutput() =
        recipeOutputDto(
            this,
            recipeImageService.getAllImages(this)
                .map { this.imageUrl(it) }
                .toSet()
        )

    private fun Recipe.imageUrl(name: String) =
        "${creProperties.deploymentUrl}$FILE_CONTROLLER_PATH?path=${
            URLEncoder.encode(
                "${this.imagesDirectoryPath}/$name",
                StandardCharsets.UTF_8
            )
        }"
}

@RestController
@RequestMapping(MIX_CONTROLLER_PATH)
@PreAuthorizeViewRecipes
class MixController(private val mixService: MixService) {
    @GetMapping("{id}")
    fun getById(@PathVariable id: Long) =
        ok(mixService.getById(id))

    @PostMapping
    @PreAuthorizeEditRecipes
    fun save(@Valid @RequestBody mix: MixSaveDto) =
        created<Mix>(MIX_CONTROLLER_PATH) {
            mixService.save(mix)
        }

    @PutMapping
    @PreAuthorizeEditRecipes
    fun update(@Valid @RequestBody mix: MixUpdateDto) =
        noContent {
            mixService.update(mix)
        }

    @DeleteMapping("{id}")
    @PreAuthorizeRemoveRecipes
    fun deleteById(@PathVariable id: Long) =
        noContent {
            mixService.deleteById(id)
        }
}
