package dev.fyloz.colorrecipesexplorer.rest

import dev.fyloz.colorrecipesexplorer.model.Model
import org.springframework.http.HttpHeaders
import org.springframework.http.HttpStatus
import org.springframework.http.MediaType
import org.springframework.http.ResponseEntity
import java.net.URI

/** Creates a HTTP OK [ResponseEntity] from the given [body]. */
fun <T> ok(body: T): ResponseEntity<T> =
    ResponseEntity.ok(body)

/** Creates a HTTP OK [ResponseEntity] from the given [body] and [headers]. */
fun <T> ok(body: T, headers: HttpHeaders): ResponseEntity<T> =
    ResponseEntity(body, headers, HttpStatus.OK)

/** Executes the given [action] then returns an HTTP OK [ResponseEntity] form the given [body]. */
fun <T> ok(action: () -> Unit): ResponseEntity<T> {
    action()
    return ResponseEntity.ok().build()
}

/** Creates a HTTP CREATED [ResponseEntity] from the given [body] with the location set to [controllerPath]/id. */
fun <T : Model> created(controllerPath: String, body: T): ResponseEntity<T> =
    created(controllerPath, body, body.id!!)

/** Creates a HTTP CREATED [ResponseEntity] with the result of the given [producer] as its body. */
fun <T : Model> created(controllerPath: String, producer: () -> T): ResponseEntity<T> =
    created(controllerPath, producer())

/** Creates a HTTP CREATED [ResponseEntity] from the given [body] with the location set to [controllerPath]/id. */
fun <T> created(controllerPath: String, body: T, id: Any): ResponseEntity<T> =
    ResponseEntity.created(URI.create("$controllerPath/$id")).body(body)

/** Creates a HTTP NOT FOUND [ResponseEntity]. */
fun <T> notFound(): ResponseEntity<T> =
    ResponseEntity.notFound().build()

/** Creates a HTTP NO CONTENT [ResponseEntity]. */
fun noContent(): ResponseEntity<Void> =
    ResponseEntity.noContent().build()

/** Executes the given [action] then returns an HTTP NO CONTENT [ResponseEntity]. */
fun noContent(action: () -> Unit): ResponseEntity<Void> {
    action()
    return noContent()
}

/** Creates a HTTP FORBIDDEN [ResponseEntity]. */
fun <T> forbidden(): ResponseEntity<T> =
    ResponseEntity.status(HttpStatus.FORBIDDEN).build()

/** Creates an [HttpHeaders] instance from the given options. */
fun httpHeaders(
    contentType: MediaType = MediaType.APPLICATION_JSON,
    op: HttpHeaders.() -> Unit = {}
) = HttpHeaders().apply {
    this.contentType = contentType

    op()
}
