package dev.fyloz.colorrecipesexplorer.service.files

import dev.fyloz.colorrecipesexplorer.config.properties.CreProperties
import dev.fyloz.colorrecipesexplorer.exception.RestException
import org.slf4j.Logger
import org.springframework.core.io.ByteArrayResource
import org.springframework.http.HttpStatus
import org.springframework.stereotype.Service
import org.springframework.web.multipart.MultipartFile
import java.io.File
import java.io.IOException
import java.nio.file.Files

interface FileService {
    /** Checks if the file at the given [path] exists. */
    fun exists(path: String): Boolean

    /** Reads the file at the given [path]. */
    fun read(path: String): ByteArrayResource

    /** Creates a file at the given [path]. */
    fun create(path: String)

    /** Writes the given [file] at the given [path]. If the file already exists, it will be overwritten if [overwrite] is true. */
    fun write(file: MultipartFile, path: String, overwrite: Boolean)

    /** Deletes the file at the given [path]. */
    fun delete(path: String)

    /** Completes the path of the given [String] by adding the working directory. */
    fun String.fullPath(): FilePath
}

@Service
class FileServiceImpl(
    private val creProperties: CreProperties,
    private val logger: Logger
) : FileService {
    override fun exists(path: String) = withFileAt(path.fullPath()) {
        this.exists() && this.isFile
    }

    override fun read(path: String) = ByteArrayResource(
        withFileAt(path.fullPath()) {
            if (!exists(path)) throw FileNotFoundException(path)
            try {
                readBytes()
            } catch (ex: IOException) {
                FileReadException(path).logAndThrow(ex, logger)
            }
        }
    )

    override fun create(path: String) {
        val fullPath = path.fullPath()
        if (!exists(path)) {
            try {
                withFileAt(fullPath) {
                    this.create()
                }
            } catch (ex: IOException) {
                FileCreateException(path).logAndThrow(ex, logger)
            }
        }
    }

    override fun write(file: MultipartFile, path: String, overwrite: Boolean) {
        val fullPath = path.fullPath()

        if (exists(path)) {
            if (!overwrite) throw FileExistsException(path)
        } else {
            create(path)
        }

        try {
            withFileAt(fullPath) {
                file.transferTo(this.toPath())
            }
        } catch (ex: IOException) {
            FileWriteException(path).logAndThrow(ex, logger)
        }
    }

    override fun delete(path: String) {
        try {
            withFileAt(path.fullPath()) {
                if (!exists(path)) throw FileNotFoundException(path)
                !this.delete()
            }
        } catch (ex: IOException) {
            FileDeleteException(path).logAndThrow(ex, logger)
        }
    }

    override fun String.fullPath() =
        FilePath("${creProperties.workingDirectory}/$this")

    /** Runs the given [block] in the context of a file with the given [fullPath]. */
    private fun <T> withFileAt(fullPath: FilePath, block: File.() -> T) =
        fullPath.file.block()
}

data class FilePath(val path: String) {
    val file: File
        get() = File(path)
}

/** Shortcut to create a file and its parent directories. */
fun File.create() {
    Files.createDirectories(this.parentFile.toPath())
    Files.createFile(this.toPath())
}

private const val FILE_IO_EXCEPTION_TITLE = "File IO error"

class FileExistsException(val path: String) :
    RestException(
        "io-exists",
        FILE_IO_EXCEPTION_TITLE,
        HttpStatus.BAD_REQUEST,
        "Could not write file to '$path' because it already exists. To overwrite the file set the overwrite parameter to true",
        pathMap(path)
    )

class FileNotFoundException(val path: String) :
    RestException(
        "io-notfound",
        FILE_IO_EXCEPTION_TITLE,
        HttpStatus.NOT_FOUND,
        "Could not access file at '$path' because it does not exists",
        pathMap(path)
    )

sealed class FileIOException(type: String, details: String, val path: String) :
    RestException(
        "io-$type",
        FILE_IO_EXCEPTION_TITLE,
        HttpStatus.INTERNAL_SERVER_ERROR,
        details,
        pathMap(path)
    )

class FileReadException(path: String) :
    FileIOException(
        "read",
        "Could not read file at '$path'",
        path
    )

class FileWriteException(path: String) :
    FileIOException(
        "write",
        "Could not write file to '$path'",
        path
    )

class FileCreateException(path: String) :
    FileIOException(
        "create",
        "Could not create file at '$path'",
        path
    )

class FileDeleteException(path: String) :
    FileIOException(
        "delete",
        "Could not delete file at '$path'",
        path
    )

private fun pathMap(path: String) =
    mapOf("path" to path)

private fun <T : FileIOException> T.logAndThrow(baseException: Exception, logger: Logger): Nothing {
    logger.error(this.details, baseException)
    throw this
}
